@@ -1,5 +1,7 @@
 from __future__ import annotations
 
+import logging
+import os
 from types import TracebackType
 from typing import Any, Callable, Coroutine, Optional, Type
 
@@ -12,6 +14,8 @@ from linked_services.core.settings import get_setting
 
 __all__ = ["Service"]
 
+DEBUG = os.getenv("LOG_LEVEL") == "DEBUG"
+
 LIBRARIES = {
     "requests": False,
     "aiohttp": False,  # no implemented yet
@@ -94,6 +98,13 @@ if LIBRARIES["requests"]:
                 for header in header_keys:
                     resource[header] = response.headers[header]
 
+                if DEBUG:
+                    print("Response")
+                    print("  Content: no visible due to it's a stream")
+                    print("  Headers: " + str(response.headers))
+                    print("  Status code: " + str(response.status_code))
+                    print("")
+
                 return resource
 
             headers = {}
@@ -101,6 +112,14 @@ if LIBRARIES["requests"]:
             for header in header_keys:
                 headers[header] = response.headers[header]
 
+            if DEBUG:
+                print("Response")
+                print("  Type: Proxy")
+                print("  Content: " + response.content.decode())
+                print("  Headers: " + str(headers))
+                print("  Status code: " + str(response.status_code))
+                print("")
+
             return HttpResponse(response.content, status=response.status_code, headers=headers)
 
         def _sync_get(self, url, params=None, **kwargs):
@@ -112,6 +131,12 @@ if LIBRARIES["requests"]:
             headers = self._authenticate("get", params=params, **kwargs)
 
             def request() -> requests.Response:
+                if DEBUG:
+                    print("Request")
+                    print("  Method: GET")
+                    print("  Url: " + str(url))
+                    print("")
+
                 return requests.get(url, params=params, **kwargs, headers=headers)
 
             if self.proxy:
@@ -126,6 +151,12 @@ if LIBRARIES["requests"]:
             headers = self._authenticate("options", **kwargs)
 
             def request() -> requests.Response:
+                if DEBUG:
+                    print("Request")
+                    print("  Method: OPTIONS")
+                    print("  Url: " + str(url))
+                    print("")
+
                 return requests.options(url, **kwargs, headers=headers)
 
             if self.proxy:
@@ -140,6 +171,12 @@ if LIBRARIES["requests"]:
             headers = self._authenticate("head", **kwargs)
 
             def request() -> requests.Response:
+                if DEBUG:
+                    print("Request")
+                    print("  Method: HEAD")
+                    print("  Url: " + str(url))
+                    print("")
+
                 return requests.head(url, **kwargs, headers=headers)
 
             if self.proxy:
@@ -154,6 +191,12 @@ if LIBRARIES["requests"]:
             headers = self._authenticate("post", data=data, json=json, **kwargs)
 
             def request() -> requests.Response:
+                if DEBUG:
+                    print("Request")
+                    print("  Method: POST")
+                    print("  Url: " + str(url))
+                    print("")
+
                 return requests.post(url, data=data, json=json, **kwargs, headers=headers)
 
             if self.proxy:
@@ -168,6 +211,12 @@ if LIBRARIES["requests"]:
             headers = self._authenticate("post", data=data, json=json, **kwargs)
 
             def request() -> requests.Response:
+                if DEBUG:
+                    print("Request")
+                    print("  Type: Webhook")
+                    print("  Url: " + str(url))
+                    print("")
+
                 return requests.post(url, data=data, json=json, **kwargs, headers=headers)
 
             if self.proxy:
@@ -182,6 +231,12 @@ if LIBRARIES["requests"]:
             headers = self._authenticate("put", data=data, **kwargs)
 
             def request() -> requests.Response:
+                if DEBUG:
+                    print("Request")
+                    print("  Method: PUT")
+                    print("  Url: " + str(url))
+                    print("")
+
                 return requests.put(url, data=data, **kwargs, headers=headers)
 
             if self.proxy:
@@ -196,6 +251,12 @@ if LIBRARIES["requests"]:
             headers = self._authenticate("patch", data=data, **kwargs)
 
             def request() -> requests.Response:
+                if DEBUG:
+                    print("Request")
+                    print("  Method: PATCH")
+                    print("  Url: " + str(url))
+                    print("")
+
                 return requests.patch(url, data=data, **kwargs, headers=headers)
 
             if self.proxy:
@@ -210,6 +271,12 @@ if LIBRARIES["requests"]:
             headers = self._authenticate("delete", **kwargs)
 
             def request() -> requests.Response:
+                if DEBUG:
+                    print("Request")
+                    print("  Method: DELETE")
+                    print("  Url: " + str(url))
+                    print("")
+
                 return requests.delete(url, **kwargs, headers=headers)
 
             if self.proxy:
@@ -224,6 +291,12 @@ if LIBRARIES["requests"]:
             headers = self._authenticate(method, **kwargs)
 
             def request() -> requests.Response:
+                if DEBUG:
+                    print("Request")
+                    print("  Method: " + str(method).upper())
+                    print("  Url: " + str(url))
+                    print("")
+
                 return requests.request(method, url, **kwargs, headers=headers)
 
             if self.proxy:
@@ -242,6 +315,7 @@ else:
 if LIBRARIES["aiohttp"]:
 
     class AsyncService:
+        to_close: list
 
         async def __aenter__(self) -> "Service":
 
@@ -262,9 +336,17 @@ if LIBRARIES["aiohttp"]:
 
         async def __aexit__(self, *args, **kwargs) -> None:
             for obj in self.to_close:
-                await obj.__aexit__(*args, **kwargs)
+                try:
+                    await obj.__aexit__(*args, **kwargs)
 
-            await self.session.__aexit__(*args, **kwargs)
+                except Exception:
+                    pass
+
+            try:
+                await self.session.__aexit__(*args, **kwargs)
+
+            except Exception:
+                pass
 
         # django does not support StreamingHttpResponse with aiohttp due to django would have to close the response
         async def _async_proxy(self, response: Coroutine[Any, Any, ClientResponse]) -> HttpResponse:
@@ -280,7 +362,17 @@ if LIBRARIES["aiohttp"]:
             for header in header_keys:
                 headers[str(header)] = r.headers[header]
 
-            return HttpResponse(await r.content.read(), status=r.status, headers=headers)
+            content = await r.content.read()
+
+            if DEBUG:
+                print("Response")
+                print("  Type: Proxy")
+                print("  Content: " + content.decode())
+                print("  Headers: " + str(headers))
+                print("  Status code: " + str(r.status))
+                print("")
+
+            return HttpResponse(content, status=r.status, headers=headers)
 
         def _async_get(self, url, params=None, **kwargs):
             url = self.app.app_url + self._fix_url(url)
@@ -291,6 +383,12 @@ if LIBRARIES["aiohttp"]:
 
             res = obj.__aenter__()
 
+            if DEBUG:
+                print("Request")
+                print("  Method: GET")
+                print("  Url: " + str(url))
+                print("")
+
             # wraps client response to be used within django views
             if self.proxy:
                 return self._async_proxy(res)
@@ -306,6 +404,12 @@ if LIBRARIES["aiohttp"]:
 
             res = obj.__aenter__()
 
+            if DEBUG:
+                print("Request")
+                print("  Method: OPTIONS")
+                print("  Url: " + str(url))
+                print("")
+
             # wraps client response to be used within django views
             if self.proxy:
                 return self._async_proxy(res)
@@ -321,6 +425,12 @@ if LIBRARIES["aiohttp"]:
 
             res = obj.__aenter__()
 
+            if DEBUG:
+                print("Request")
+                print("  Method: HEAD")
+                print("  Url: " + str(url))
+                print("")
+
             # wraps client response to be used within django views
             if self.proxy:
                 return self._async_proxy(res)
@@ -336,6 +446,12 @@ if LIBRARIES["aiohttp"]:
 
             res = obj.__aenter__()
 
+            if DEBUG:
+                print("Request")
+                print("  Method: POST")
+                print("  Url: " + str(url))
+                print("")
+
             # wraps client response to be used within django views
             if self.proxy:
                 return self._async_proxy(res)
@@ -351,6 +467,12 @@ if LIBRARIES["aiohttp"]:
 
             res = obj.__aenter__()
 
+            if DEBUG:
+                print("Request")
+                print("  Type: Webhook")
+                print("  Url: " + str(url))
+                print("")
+
             # wraps client response to be used within django views
             if self.proxy:
                 return self._async_proxy(res)
@@ -366,6 +488,12 @@ if LIBRARIES["aiohttp"]:
 
             res = obj.__aenter__()
 
+            if DEBUG:
+                print("Request")
+                print("  Method: PUT")
+                print("  Url: " + str(url))
+                print("")
+
             # wraps client response to be used within django views
             if self.proxy:
                 return self._async_proxy(res)
@@ -381,6 +509,12 @@ if LIBRARIES["aiohttp"]:
 
             res = obj.__aenter__()
 
+            if DEBUG:
+                print("Request")
+                print("  Method: PATCH")
+                print("  Url: " + str(url))
+                print("")
+
             # wraps client response to be used within django views
             if self.proxy:
                 return self._async_proxy(res)
@@ -396,6 +530,12 @@ if LIBRARIES["aiohttp"]:
 
             res = obj.__aenter__()
 
+            if DEBUG:
+                print("Request")
+                print("  Method: DELETE")
+                print("  Url: " + str(url))
+                print("")
+
             # wraps client response to be used within django views
             if self.proxy:
                 return self._async_proxy(res)
@@ -411,6 +551,12 @@ if LIBRARIES["aiohttp"]:
 
             res = obj.__aenter__()
 
+            if DEBUG:
+                print("Request")
+                print("  Method: " + str(method).upper())
+                print("  Url: " + str(url))
+                print("")
+
             # wraps client response to be used within django views
             if self.proxy:
                 return self._async_proxy(res)
@@ -449,6 +595,13 @@ elif LIBRARIES["requests"]:
                 for header in header_keys:
                     resource[header] = response.headers[header]
 
+                if DEBUG:
+                    print("Response")
+                    print("  Content: no visible due to it's a stream")
+                    print("  Headers: " + str(response.headers))
+                    print("  Status code: " + str(response.status_code))
+                    print("")
+
                 return resource
 
             headers = {}
@@ -456,6 +609,14 @@ elif LIBRARIES["requests"]:
             for header in header_keys:
                 headers[header] = response.headers[header]
 
+            if DEBUG:
+                print("Response")
+                print("  Type: Proxy")
+                print("  Content: " + response.content.decode())
+                print("  Headers: " + str(headers))
+                print("  Status code: " + str(response.status_code))
+                print("")
+
             return HttpResponse(response.content, status=response.status_code, headers=headers)
 
         async def _async_get(self, url, params=None, **kwargs):
